@@ -29,6 +29,8 @@
 
 #include "connman.h"
 
+#define CONNECT_TIMEOUT		120
+
 static DBusConnection *connection = NULL;
 
 static GSequence *service_list = NULL;
@@ -54,6 +56,12 @@ struct connman_service {
 	char *name;
 	char *passphrase;
 	char *profile;
+	char *apn;
+	char *username;
+	char *password;
+	char *mcc;
+	char *mnc;
+	connman_bool_t roaming;
 	struct connman_ipconfig *ipconfig;
 	struct connman_device *device;
 	struct connman_network *network;
@@ -137,6 +145,12 @@ static const char *mode2string(enum connman_service_mode mode)
 		return "managed";
 	case CONNMAN_SERVICE_MODE_ADHOC:
 		return "adhoc";
+	case CONNMAN_SERVICE_MODE_GPRS:
+		return "gprs";
+	case CONNMAN_SERVICE_MODE_EDGE:
+		return "edge";
+	case CONNMAN_SERVICE_MODE_UMTS:
+		return "umts";
 	}
 
 	return NULL;
@@ -232,6 +246,36 @@ const char *__connman_service_default(void)
 	return type2string(service->type);
 }
 
+static void mode_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *str, *key = "Mode";
+
+	if (service->path == NULL)
+		return;
+
+	str = mode2string(service->mode);
+	if (str == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_STRING_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &str);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void state_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -292,6 +336,60 @@ static void strength_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void roaming_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Roaming";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->roaming);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static void autoconnect_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "AutoConnect";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->autoconnect);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void passphrase_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -343,15 +441,29 @@ static void passphrase_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
-static void autoconnect_changed(struct connman_service *service)
+static void apn_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
 	DBusMessageIter entry, value;
-	const char *key = "AutoConnect";
+	dbus_bool_t required;
+	const char *key = "SetupRequired";
 
 	if (service->path == NULL)
 		return;
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		return;
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		break;
+	}
+
+	required = (service->apn == NULL) ? TRUE : FALSE;
+
 	signal = dbus_message_new_signal(service->path,
 				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
 	if (signal == NULL)
@@ -363,8 +475,7 @@ static void autoconnect_changed(struct connman_service *service)
 
 	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
 					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
-							&service->autoconnect);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &required);
 	dbus_message_iter_close_container(&entry, &value);
 
 	g_dbus_send_message(connection, signal);
@@ -440,7 +551,38 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		connman_dbus_dict_append_variant(&dict, "Roaming",
+					DBUS_TYPE_BOOLEAN, &service->roaming);
+
+		if (service->mcc != NULL && service->mnc != NULL) {
+			connman_dbus_dict_append_variant(&dict, "MCC",
+					DBUS_TYPE_STRING, &service->mcc);
+			connman_dbus_dict_append_variant(&dict, "MNC",
+					DBUS_TYPE_STRING, &service->mnc);
+		}
+
+		if (service->apn != NULL) {
+			connman_dbus_dict_append_variant(&dict, "APN",
+					DBUS_TYPE_STRING, &service->apn);
+
+			if (service->username != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Username", DBUS_TYPE_STRING,
+							&service->username);
+
+			if (service->password != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Password", DBUS_TYPE_STRING,
+							&service->password);
+
+			required = FALSE;
+		} else
+			required = TRUE;
+
+		connman_dbus_dict_append_variant(&dict, "SetupRequired",
+						DBUS_TYPE_BOOLEAN, &required);
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		if (service->passphrase != NULL &&
@@ -500,7 +642,26 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	type = dbus_message_iter_get_arg_type(&value);
 
-	if (g_str_equal(name, "Passphrase") == TRUE) {
+	if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
+		connman_bool_t autoconnect;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->favorite == FALSE)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &autoconnect);
+
+		if (service->autoconnect == autoconnect)
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		service->autoconnect = autoconnect;
+
+		autoconnect_changed(service);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Passphrase") == TRUE) {
 		const char *passphrase;
 
 		if (type != DBUS_TYPE_STRING)
@@ -522,23 +683,63 @@ static DBusMessage *set_property(DBusConnection *conn,
 				"WiFi.Passphrase", service->passphrase);
 
 		__connman_storage_save_service(service);
-	} else if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
-		connman_bool_t autoconnect;
+	} else if (g_str_equal(name, "APN") == TRUE) {
+		const char *apn;
 
-		if (type != DBUS_TYPE_BOOLEAN)
+		if (type != DBUS_TYPE_STRING)
 			return __connman_error_invalid_arguments(msg);
 
-		if (service->favorite == FALSE)
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
 			return __connman_error_invalid_service(msg);
 
-		dbus_message_iter_get_basic(&value, &autoconnect);
+		dbus_message_iter_get_basic(&value, &apn);
 
-		if (service->autoconnect == autoconnect)
-			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+		g_free(service->apn);
+		service->apn = g_strdup(apn);
 
-		service->autoconnect = autoconnect;
+		apn_changed(service);
 
-		autoconnect_changed(service);
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Username") == TRUE) {
+		const char *username;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &username);
+
+		g_free(service->username);
+		service->username = g_strdup(username);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Password") == TRUE) {
+		const char *password;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &password);
+
+		g_free(service->password);
+		service->password = g_strdup(password);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
 
 		__connman_storage_save_service(service);
 	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
@@ -620,6 +821,9 @@ static connman_bool_t is_ignore(struct connman_service *service)
 	if (service->autoconnect == FALSE)
 		return TRUE;
 
+	if (service->roaming == TRUE)
+		return TRUE;
+
 	if (service->ignore == TRUE)
 		return TRUE;
 
@@ -668,12 +872,17 @@ void __connman_service_auto_connect(void)
 	}
 }
 
-static void reply_pending(struct connman_service *service, int error)
+static void remove_timeout(struct connman_service *service)
 {
 	if (service->timeout > 0) {
 		g_source_remove(service->timeout);
 		service->timeout = 0;
 	}
+}
+
+static void reply_pending(struct connman_service *service, int error)
+{
+	remove_timeout(service);
 
 	if (service->pending != NULL) {
 		if (error > 0) {
@@ -729,6 +938,40 @@ static gboolean connect_timeout(gpointer user_data)
 	return FALSE;
 }
 
+static void set_reconnect_state(struct connman_service *service,
+						connman_bool_t reconnect)
+{
+	if (service->network != NULL) {
+		struct connman_device *device;
+
+		device = connman_network_get_device(service->network);
+
+		__connman_device_set_reconnect(device, reconnect);
+
+		return;
+	}
+
+	if (service->device != NULL)
+		__connman_device_set_reconnect(service->device,
+							reconnect);
+}
+
+static connman_bool_t get_reconnect_state(struct connman_service *service)
+{
+	if (service->network != NULL) {
+		struct connman_device *device;
+
+		device = connman_network_get_device(service->network);
+
+		return __connman_device_get_reconnect(device);
+	}
+
+	if (service->device != NULL)
+		return __connman_device_get_reconnect(service->device);
+
+	return FALSE;
+}
+
 static DBusMessage *connect_service(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -759,6 +1002,8 @@ static DBusMessage *connect_service(DBusConnection *conn,
 
 	service->pending = dbus_message_ref(msg);
 
+	set_reconnect_state(service, FALSE);
+
 	err = __connman_service_connect(service);
 	if (err < 0) {
 		if (err == -ENOKEY) {
@@ -791,8 +1036,12 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 
 	DBG("service %p", service);
 
+	reply_pending(service, ECONNABORTED);
+
 	service->ignore = TRUE;
 
+	set_reconnect_state(service, FALSE);
+
 	err = __connman_service_disconnect(service);
 	if (err < 0) {
 		if (err != -EINPROGRESS)
@@ -817,14 +1066,28 @@ static DBusMessage *remove_service(DBusConnection *conn,
 	if (service->favorite == FALSE)
 		return __connman_error_not_supported(msg);
 
-	if (service->network != NULL)
+	if (service->network != NULL) {
+		set_reconnect_state(service, FALSE);
+
 		__connman_network_disconnect(service->network);
+	}
 
 	g_free(service->passphrase);
 	service->passphrase = NULL;
 
 	passphrase_changed(service);
 
+	g_free(service->apn);
+	service->apn = NULL;
+
+	g_free(service->username);
+	service->username = NULL;
+
+	g_free(service->password);
+	service->password = NULL;
+
+	apn_changed(service);
+
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
@@ -946,6 +1209,11 @@ static void service_free(gpointer user_data)
 		service->ipconfig = NULL;
 	}
 
+	g_free(service->mcc);
+	g_free(service->mnc);
+	g_free(service->apn);
+	g_free(service->username);
+	g_free(service->password);
 	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
@@ -968,7 +1236,10 @@ void __connman_service_put(struct connman_service *service)
 
 		iter = g_hash_table_lookup(service_hash, service->identifier);
 		if (iter != NULL) {
+			reply_pending(service, ECONNABORTED);
+
 			__connman_service_disconnect(service);
+
 			g_sequence_remove(iter);
 		} else
 			service_free(service);
@@ -1084,6 +1355,33 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 	return (gint) service_b->strength - (gint) service_a->strength;
 }
 
+static void favorite_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Favorite";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->favorite);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 /**
  * connman_service_set_favorite:
  * @service: service structure
@@ -1105,6 +1403,8 @@ int connman_service_set_favorite(struct connman_service *service,
 
 	service->favorite = favorite;
 
+	favorite_changed(service);
+
 	g_sequence_sort_changed(iter, service_compare, NULL);
 
 	__connman_profile_changed(FALSE);
@@ -1162,13 +1462,25 @@ int __connman_service_indicate_state(struct connman_service *service,
 		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
 		state_changed(service);
 
+		reply_pending(service, ECONNABORTED);
+
 		__connman_service_disconnect(service);
 	}
 
 	service->state = state;
 	state_changed(service);
 
+	if (state == CONNMAN_SERVICE_STATE_IDLE) {
+		connman_bool_t reconnect;
+
+		reconnect = get_reconnect_state(service);
+		if (reconnect == TRUE)
+			__connman_service_auto_connect();
+	}
+
 	if (state == CONNMAN_SERVICE_STATE_READY) {
+		set_reconnect_state(service, TRUE);
+
 		connman_service_set_favorite(service, TRUE);
 
 		reply_pending(service, 0);
@@ -1256,8 +1568,17 @@ static connman_bool_t prepare_network(struct connman_service *service)
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_CELLULAR:
+		break;
 	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
+		connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+		connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
 		break;
 	}
 
@@ -1282,7 +1603,10 @@ int __connman_service_connect(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn == NULL)
+			return -EINVAL;
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		switch (service->security) {
@@ -1322,7 +1646,7 @@ int __connman_service_connect(struct connman_service *service)
 			return err;
 		}
 
-		service->timeout = g_timeout_add_seconds(45,
+		service->timeout = g_timeout_add_seconds(CONNECT_TIMEOUT,
 						connect_timeout, service);
 
 		return -EINPROGRESS;
@@ -1337,8 +1661,6 @@ int __connman_service_disconnect(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	reply_pending(service, ECONNABORTED);
-
 	if (service->network != NULL) {
 		err = __connman_network_disconnect(service->network);
 	} else if (service->device != NULL) {
@@ -1535,6 +1857,8 @@ done:
 		goto failed;
 	}
 
+	set_reconnect_state(service, FALSE);
+
 	__connman_device_disconnect(device);
 
 	if (passphrase != NULL) {
@@ -1860,6 +2184,7 @@ static enum connman_service_type convert_network_type(struct connman_network *ne
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
 		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+	case CONNMAN_NETWORK_TYPE_CELLULAR:
 	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
 		return CONNMAN_SERVICE_TYPE_CELLULAR;
@@ -1896,6 +2221,24 @@ static enum connman_service_mode convert_wifi_security(const char *security)
 		return CONNMAN_SERVICE_SECURITY_UNKNOWN;
 }
 
+static enum connman_service_mode convert_cellular_mode(connman_uint8_t mode)
+{
+	switch (mode) {
+	case 0:
+	case 1:
+		return CONNMAN_SERVICE_MODE_GPRS;
+	case 3:
+		return CONNMAN_SERVICE_MODE_EDGE;
+	case 2:
+	case 4:
+	case 5:
+	case 6:
+		return CONNMAN_SERVICE_MODE_UMTS;
+	}
+
+	return CONNMAN_SERVICE_MODE_UNKNOWN;
+}
+
 static void update_from_network(struct connman_service *service,
 					struct connman_network *network)
 {
@@ -1921,6 +2264,15 @@ static void update_from_network(struct connman_service *service,
 	}
 
 	service->strength = connman_network_get_uint8(network, "Strength");
+	service->roaming = connman_network_get_bool(network, "Roaming");
+
+	if (service->strength == 0) {
+		/*
+		 * Filter out 0-values; it's unclear what they mean
+		 * and they cause anomalous sorting of the priority list.
+		 */
+		service->strength = strength;
+	}
 
 	str = connman_network_get_string(network, "WiFi.Mode");
 	service->mode = convert_wifi_mode(str);
@@ -1928,6 +2280,21 @@ static void update_from_network(struct connman_service *service,
 	str = connman_network_get_string(network, "WiFi.Security");
 	service->security = convert_wifi_security(str);
 
+	str = connman_network_get_string(network, "Cellular.MCC");
+	g_free(service->mcc);
+	service->mcc = g_strdup(str);
+
+	str = connman_network_get_string(network, "Cellular.MNC");
+	g_free(service->mnc);
+	service->mnc = g_strdup(str);
+
+	if (service->type == CONNMAN_SERVICE_TYPE_CELLULAR) {
+		connman_uint8_t value = connman_network_get_uint8(network,
+							"Cellular.Mode");
+
+		service->mode = convert_cellular_mode(value);
+	}
+
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
 		service->network = connman_network_ref(network);
@@ -1987,10 +2354,10 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		service->autoconnect = FALSE;
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		service->autoconnect = TRUE;
 		break;
 	}
@@ -2014,7 +2381,10 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 void __connman_service_update_from_network(struct connman_network *network)
 {
 	struct connman_service *service;
-	connman_uint8_t strength;
+	enum connman_service_mode mode;
+	connman_uint8_t strength, value;
+	connman_bool_t roaming;
+	GSequenceIter *iter;
 
 	service = __connman_service_lookup_from_network(network);
 	if (service == NULL)
@@ -2025,11 +2395,38 @@ void __connman_service_update_from_network(struct connman_network *network)
 
 	strength = connman_network_get_uint8(service->network, "Strength");
 	if (strength == service->strength)
-		return;
+		goto roaming;
 
 	service->strength = strength;
 
 	strength_changed(service);
+
+roaming:
+	roaming = connman_network_get_bool(service->network, "Roaming");
+	if (roaming == service->roaming)
+		goto done;
+
+	service->roaming = roaming;
+
+	roaming_changed(service);
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL)
+		g_sequence_sort_changed(iter, service_compare, NULL);
+
+done:
+	if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+		return;
+
+	value = connman_network_get_uint8(service->network, "Cellular.Mode");
+	mode = convert_cellular_mode(value);
+
+	if (mode == service->mode)
+		return;
+
+	service->mode = mode;
+
+	mode_changed(service);
 }
 
 void __connman_service_remove_from_network(struct connman_network *network)
@@ -2138,6 +2535,15 @@ static int service_load(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->apn = g_key_file_get_string(keyfile,
+					service->identifier, "APN", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Username", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Password", NULL);
+
 		service->favorite = g_key_file_get_boolean(keyfile,
 				service->identifier, "Favorite", NULL);
 
@@ -2254,6 +2660,18 @@ update:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+							"APN", service->apn);
+
+		if (service->username != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Username", service->username);
+
+		if (service->password != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Password", service->password);
+
 		g_key_file_set_boolean(keyfile, service->identifier,
 					"Favorite", service->favorite);
 
